/*
 * Copyright 2002-2016 jamod & j2mod development teams
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.ghgande.j2mod.modbus.msg;

import com.ghgande.j2mod.modbus.Modbus;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;

/**
 * Class implementing a<tt>ModbusResponse</tt> that represents an exception.
 *
 * @author Dieter Wimberger
 * @author Steve O'Hara (4NG)
 * @version 2.0 (March 2016)
 */
public class ExceptionResponse extends ModbusResponse {

    // instance attributes
    private int exceptionCode = -1;

    /**
     * Constructs a new <tt>ExceptionResponse</tt> instance with a given
     * function code and an exception code. The function code will be
     * automatically increased with the exception offset.
     *
     * @param fc  the function code as <tt>int</tt>.
     * @param exc the exception code as <tt>int</tt>.
     */
    public ExceptionResponse(int fc, int exc) {

        // One byte of data.
        setDataLength(1);
        setFunctionCode(fc | Modbus.EXCEPTION_OFFSET);

        exceptionCode = exc;
    }

    /**
     * Constructs a new <tt>ExceptionResponse</tt> instance with a given
     * function code. ORs the exception offset automatically.
     *
     * @param fc the function code as <tt>int</tt>.
     */
    public ExceptionResponse(int fc) {

        // One byte of data.
        setDataLength(1);
        setFunctionCode(fc | Modbus.EXCEPTION_OFFSET);
    }

    /**
     * Constructs a new <tt>ExceptionResponse</tt> instance with no function
     * or exception code.
     */
    public ExceptionResponse() {

        // One byte of data.
        setDataLength(1);
    }

    /**
     * Returns the Modbus exception code of this <tt>ExceptionResponse</tt>.
     * <p>
     *
     * @return the exception code as <tt>int</tt>.
     */
    public int getExceptionCode() {
        return exceptionCode;
    }

    public void writeData(DataOutput dout) throws IOException {
        dout.writeByte(getExceptionCode());
    }

    /**
     * readData()
     *
     * read the single byte of data, which is the exception code.
     * @throws IOException If the data cannot be read from the socket/port
     */
    public void readData(DataInput din) throws IOException {
        exceptionCode = din.readUnsignedByte();
    }

    /**
     * getMessage()
     *
     * return the exception type, which is the "message" for this response.
     *
     * @return -- byte array containing the 1 byte exception code.
     */
    public byte[] getMessage() {
        byte result[] = new byte[1];
        result[0] = (byte)getExceptionCode();
        return result;
    }
}
